#!/usr/bin/env python
# -*- mode:Python; tab-width: 4 -*-
from __future__ import generators
"""Make a local snapshot from a web tree.

cms_freeze is a tool which can freeze a given web tree generated
dynamically through OpenACS. This involves saving all files to the
local disk with correct file extensions and transforming absolute
links to relative ones, so the snapshot can be viewed offline.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""

__author__ = "Grzegorz Adam Hankiewicz <gradha@efaber.net>"
__date__ = "May 2003"
__version__ = "$Revision: 1.66 $"
__credits__ = ""

import os
import urlparse
import urllib2
import time
import re
import sys
import html_parser
import css_parser

# a few constant values
hr = "*" * 75
SUCCESSFUL_EXIT = 0
NO_CONTENT_TYPE_ERROR = 2
HTML_PARSE_ERROR = 3
HUMAN_VERSION = "1.0.3"


class Log_dict(dict):
    """Class to help with logging. Behaves like a normal dict except
    in the assignation of items, which are cummulative."""
    def __init__(self, *args):
        super(Log_dict, self).__init__(self, *args)

    def __setitem__(self, key, value):
        """Appends value to the list indexed by key.

        The value assignment adds value to a list indexed by key.
        If there is no list, it will be created empty. Besides,
        value is expected to be a sequence, its first element will
        be printed on screen.
        """
        print value[0]
        super(Log_dict, self).setdefault(key, []).append(value)


def is_url_out_of_scope(url, scope_filters):
    """is_url_out_of_scope(to_download_url, [filter_urls]) -> boolean

    Splits both url and scope_filter into individual components
    and goes through them comparing. Returns false when the domain
    and remote paths (up to the one extracted from scope_filter)
    match.  Remote paths are compared at the string level, not
    at the logical level (ie: splitting path in directories and
    matching them individually).  Use example:

    f("http://www.efaber.net/index", ["http://www.efaber.net/"]) -> 0
    f("http://www.efaber.net/index", ["http://www.efaber.net/urf/"]) -> 1
    f("http://www.efaber.net/index", ["http://www.efaber.net/help"]) -> 1
    """
    source = urlparse.urlparse(url)
    for scope in scope_filters:
        filter = urlparse.urlparse(scope)
        if source[1] == filter[1] and source[2][:len(filter[2])] == filter[2]:
            return 0
    return 1


def open_network_connection(url, referer_url = None, lang_code = ""):
    """Opens a net channel to retrieve the given url.

    url is what you want to download. Optionally, referer_url
    is the page you are coming from (this will set a language
    preference Cookie as well as send a correct Accept-Language
    header. lang_code is the preferred mode, set empty to retrieve
    whatever the sender wants to send.

    Returns a tuple with a file like object with the opened
    connection and the real url (will be different than the one you
    passed if you were redirected). Can Throw IOError and OSError
    with failures.
    """
    print "Opening connection to", url
    request = urllib2.Request(url)
    if lang_code:
        request.add_header("Cookie", "language_preference=%s" % lang_code)
        request.add_header("Accept-Language", "%s" % lang_code)
    if referer_url:
        request.add_header("Referer", referer_url)
    url_file = urllib2.urlopen(request)
    return (url_file, url_file.geturl())


def download_url(url, html_files, visited_urls, output_dir, remote_filters,
                 clean_html, logs, lang_code, black_list_regex,
                 referer_url = None):
    """Tries to download a file, doing quite a lot of magic.

    url is the absolute URL you want to download.

    html_files and visited_urls are both dictionaries.

    visited_urls is used to store the URLs this function has
    already visited to avoid redownloading the same thing. The key
    is the URL, and the value is the path to the local file after
    the download.

    html_files similar to visited_urls, but it only contains HTML
    files, and the mapping is the other way round: the key is the
    local file, and the value is the absolute URL the local file
    was downloaded from.

    output_dir is a path to a directory. Note that the path will
    be treated like a normal file path, which means it has to end
    in a slash, or the last component will be stripped.

    remote_filters is a list of URL preffixes. Before a file is
    downloaded, it is checked against those, and if it doesn't
    match at least one, the file is discarded and not downloaded.

    clean_html is a boolean parameter passed to the HTML parser. If
    it's true, the code will retry the parsing even with errors, but
    the second time the data will be fed first to an HTML validator,
    which modifies the final HTML code, but guarantees success in
    the parsing operation.

    logs is a Log_dict object where messages will be put.

    lang_code is the ISO 639 language code letter used to set
    language cookie before getting any file. It can be an empty
    string, meaning that you don't care about language.

    black_list_regex, see the description of download_cms().

    referer_url is the URL where you found url. It can be none, but
    it has to be a correct value for automatic redirectors which need
    to know where you are coming from. Otherwise, you will end in
    an infinite loop (ie: raised exception, program termination).

    The function returns a list of pairs with new files which should
    be downloaded.  The format of the pairs is (url, referer_url).
    """
    # remove fragment part of URL (ie: blah#fragment)
    url, dummy = html_parser.split_url(url)
    if url in visited_urls:
        return []

    # test against black list
    if black_list_regex and black_list_regex.search(url):
        logs["black-list"] = ("`%s' is blacklisted" % url, url)
        return []

    # throw away if we are out of scope and it's not a language switch
    if is_url_out_of_scope(url, remote_filters):
        if not html_parser.find_language_switch(url)[0]:
            logs["out-of-domain"] = ("'%s' out of filters" % url, url)
            return []

    try:
        url_file, real_url = open_network_connection(url, referer_url, lang_code)
    except (OSError, IOError), msg:
        logs["errors"] = ("Couldn't retrieve `%s' referenced from `%s'\n"
            "%s" % (url, referer_url, msg), "%s from %s" % (url, referer_url))
        return []

    local_path = html_parser.url_to_local_path(real_url, output_dir)

    new_lang_code, redirect_url = html_parser.find_language_switch(url)
    if new_lang_code:
        url_file.close()
        local_path = html_parser.rename_local_path(local_path, new_lang_code)
        logs["lang-switch"] = ("Language switch to '%s' found in '%s'.\n"
            "\t-> %s" % (new_lang_code, referer_url, local_path),
            "%s -> %s" % (referer_url, local_path))
        # add the redirected url to pending ones only if it had redirect part
        if redirect_url:
            logs["no-lang-switch"] = ("There is forced redirection in switch",
                "%s -> %s" % (url, local_path))
            message = "%s -> %s" % (url, local_path)
            visited_urls[url] = local_path
        return [(real_url, url)]
    
    try: # note the finally at the end of this try
        try:
            if url_file.info()["Content-type"].find("text/") >= 0:
                local_path = html_parser.rename_local_path(local_path, lang_code)
                if is_url_out_of_scope(real_url, remote_filters):
                    logs["outside-scope"] = ("`%s' falls outside of download"
                        "scope `%s'" % (real_url, remote_filters), real_url)
                    return []
        except KeyError:
            print hr
            print " All web pages should have type content but this doesn't!"
            print " I tried to download: `%s'" % url
            print " I found this link here: `%s'" % referer_url
            print hr
            sys.exit(NO_CONTENT_TYPE_ERROR)

        # see if we got redirected to something we know...
        if real_url in visited_urls:
            return []

        data = handle_data_download(url_file, local_path, logs)
        logs["download"] = ("`%s' downloaded" % real_url, real_url)

        # don't process if it's not htm or css
        ext = os.path.splitext(local_path)[1]
        if ext == ".css":
            html_files[local_path] = real_url
            return extract_css_links(real_url, logs, data)
        elif ext.find("htm") >= 0:
            html_files[local_path] = real_url
            return extract_html_links(real_url, clean_html, logs, data)
    finally:
        url_file.close()
        message = "url `%s' -> local_path `%s'" % (url, local_path)
        logs["visited-url"] = ("Visited %s" % message, message)
        message = "real_url `%s' -> local_path `%s'" % (real_url, local_path)
        logs["visited-url"] = ("Visited %s" % message, message)
        visited_urls[url] = local_path
        visited_urls[real_url] = local_path
    return []


def handle_data_download(url_file, local_path, logs):
    """Gets an URL and saves it to local_path(.tmp), returning its contents.

    This function retrieves url_file and stores it in local_path.tmp.
    logs is a Log_dict class to put there messages. Returns the
    data of the downloaded file.

    This function is clever enough to check if a file has already
    been downloaded, in which case it won't be downloaded again
    from the net. Instead, the current local file will be opened
    and it's data will be returned.
    """
    create_local_path(local_path)
    name = "%s.tmp" % local_path
    if os.path.isfile(name):
        logs["already-downloaded"] = (local_path, local_path)
        file = open(name, "rb")
        data = file.read()
    else:
        file = open(name, "wb")
        data = url_file.read()
        file.write(data)
        
    file.close()
    return data


def create_local_path(path, mode = 0777):
    """Creates directory tree up to `path'. mode defaults to 0777.

    Replacement function for os.makedirs. The only difference is that
    it won't raise exception if the path (or one of it's components)
    already exists.
    """
    built_path = ""
    for level in html_parser.split_all_path(os.path.dirname(path)):
        built_path = os.path.join(built_path, level)
        if not os.path.isdir(built_path):
            os.mkdir(built_path, mode)


def extract_html_links(url, clean_html, logs, contents):
    """Process HTML data and return its links.

    This is a wrapper around the HTML Link_extractor class. It parses
    contents (a string) and returns a list containing all the links
    it contains (which will be absolute always) in the pair format
    (extracted_link, url_it_comes_from).

    If clean_html is true, the local file will be first tydied from
    bad HTML avoiding a parsing exception being raised.

    url should be the real URL the data was downloaded from.

    logs is expected to be an object of class Log_dict, to which
    new log information will be appended.
    """
    try:
        l = html_parser.Link_extractor(contents, clean_html)
        logs["processed"] = ("`%s' was scaned for links." % url, url)
    except Exception, message:
        from urllib import quote
        print hr
        print " Problems parsing `%s'\n%s\n" % (url, message)
        print " Please try `http://validator.w3.org/check?uri=%s'" % quote(url)
        print hr
        sys.exit(HTML_PARSE_ERROR)
    return [(urlparse.urljoin(url, x), url) for x in l.get_links()]


def extract_css_links(url, logs, contents):
    """Process CSS data and return its links.

    This is a wrapper around the CSS Link_extractor class. It parses
    contents (a string) and returns a list containing all the links
    it contains (which will be absolute always) in the pair format
    (extracted_link, url_it_comes_from).

    url should be the real URL the data was downloaded from.

    logs is expected to be an object of class Log_dict, to which
    new log information will be appended.
    """
    l = css_parser.Link_extractor(contents)
    logs["processed"] = ("`%s' was scaned for links." % url, url)
    return [(urlparse.urljoin(url, x), url) for x in l.get_links()]
    

def post_process_file(local_file, url, substitutions, clean_html):
    """Post process the downloaded file replacing its links.

    Tries to open '%s.tmp' % local_file which was downloaded
    from url and processes its links, using substitutions as a
    replacement dictionary. Once the file is processed, it will be
    saved without the '.tmp' extension and the original temporary
    file will be deleted.

    If clean_html is true, the local file will be first tidied from
    bad HTML avoiding a parsing exception being raised.
    """
    temp_file = "%s.tmp" % local_file
    print "Processing `%s' for relative links" % temp_file

    if temp_file.find(".css.tmp") > 0:
        module = css_parser
    else:
        module = html_parser

    try:
        t = module.Link_transformer(temp_file, url, substitutions, clean_html)
    except IOError, msg:
        print "Couldn't open it `%s'. Maybe it was already processed?" % msg
        return
        
    file = open(local_file, "wb")
    file.write(t.get_data())
    file.close()
    os.unlink(temp_file)


def download_cms(start_urls, output_dir, remote_filters, clean_html, lang_code,
                 black_list_regex):
    """Gets the whole remote CMS tree.

    start_urls is a list of URLs you want to use as starting point for
    the web crawling. Usually only need one, but you can also use this
    as a way to make sure specific urls get downloaded.

    output_dir is the directory path where the downloaded files will
    be saved. Note that it is a 'real path', so it has to end with
    a slash or the last component will be stripped, presumed to be
    a file name.  Using more than one slash at the end of the path
    is harmless.

    remote_filters is a list of fully qualified URLs which acts as
    prefix filters. If any processed url does not match at least
    one of remote_filters, it will fall out of scope and won't be
    downloaded. This check is enforced before downloading files,
    with the exception of start_urls, which don't have a filter.

    clean_html is a boolean value. If true, bad HTML will be filtered
    if a parsing exception is thrown. Filtering bad HTML could
    modify it so that it doesn't render any more as you expect.
    If possible, fix your original code to be valid HTML.

    lang_code is an ISO 639 language code indicating the language
    you want to set in the header 'Accept-Language' sent in HTTP
    requests.  You can pass None or the empty string to indicate
    that you want to get the default language the server will return.

    black_list_regex is a regular expression object which will be
    applied over URLs before they are downloaded. URLs with positive
    matches won't be downloaded. This parameter can be None.

    Returns in a tuple the parameters you would have to use with
    the call of generate_log().
    """
    html_files = {}
    visited_urls = {}
    logs = Log_dict()
    # make copy of valuable parameters
    start_list = start_urls[:]
    pending_urls = []
    start_time = time.time()

    # the first loop is a special case of the second, with starting urls
    while start_list:
        url = start_list.pop()
        # note that we use as referer the self url, to avoid wrong filtering
        pending_urls.extend(download_url(url, html_files, visited_urls,
            output_dir, [url], clean_html, logs, lang_code, black_list_regex))

    while pending_urls:
        url, referer_url = pending_urls.pop()
        pending_urls.extend(download_url(url, html_files, visited_urls,
            output_dir, remote_filters, clean_html, logs, lang_code,
            black_list_regex, referer_url = referer_url))

    # leave post processing until we have downloaded all the languages
    post_process_data = (html_files, visited_urls, clean_html)
    end_time = time.time()
    log_data = (logs, start_time, end_time, output_dir, start_urls, remote_filters)
    
    return (post_process_data, log_data)


def generate_log(logs, start_time, end_time, output_dir, start_urls, remote_filters):
    """Writes a log with statistics about the freeze operation.

    Called after the tree download operation has ended, will write
    a file named `cms_freeze_log.txt' in output_dir, using the data
    given as parameters.
    """
    create_local_path(output_dir)
    file = open(os.path.join(output_dir, "cms_freeze_log.txt"), "wt")
    try:
        file.write("cms_freeze log\n%s\nStart urls:\n" % hr)
        for url in start_urls: file.write("\t%s\n" % url)
        file.write("Remote filters:\n")
        for filter in remote_filters: file.write("\t%s\n" % filter)
        file.write("Started at  %s\n" % time.ctime(start_time))
        file.write("Finished at %s\n" % time.ctime(end_time))
        dif_s = end_time - start_time
        dif_m = dif_s / 60.0
        dif_h = dif_s / (60.0 * 60.0)
        file.write("Elapsed time in %d seconds, %0.2f minutes or %0.2f "
            "hours\n" % (dif_s, dif_m, dif_h))
        file.write("%s\n" % hr)

        # write a log for each logged section
        for type in logs.keys():
            file.write("Section %s:\n%s\n" % (type, "-" * 60))
            result = {}
            order = 0
            lines = []
            # extract only not duplicated lines
            for entry in logs[type]:
                try:
                    dummy = result[entry[1]]
                except KeyError:
                    result[entry[1]] = order
                order += 1
            # add to a temporary list and sort
            for key, value in result.iteritems():
                lines.append((value, key))
            lines.sort()
            for line in lines: file.write("%s\n" % line[1])
            file.write("\n\n\n")
    finally:
        file.close()
    print "Wrote logs to %s" % os.path.join(output_dir, "cms_freeze_log.txt")


def walk_local_files(extension, path, strip_length = 0):
    """Walks over the temporary files of a directory.

    path is the directory you want to travel. strip_length should
    be zero for the base case, which holds the length you have to
    erase from the full path to obtain the relative path from the
    beginning of the function call.

    extension is the string you want to match. It has to match
    exactly, which means including the extension dot.

    This function uses a generator, which returns tuples in the form
    (filename, relative path from path to filename).
    """
    if not strip_length:
        strip_length = len(os.path.join(path, ".")) - 1
          
    for item in os.listdir(path):
        full = os.path.join(path, item)
        if os.path.isdir(full):
            for val, partial in walk_local_files(extension, full, strip_length):
                yield val, partial
        elif os.path.splitext(full)[1].lower() == extension:
            yield full, full[strip_length:]


def main_process(output_dir, start_urls, remote_filters, language_codes,
                 html_extension, clean_html, black_list):
    """Execute the download tasks.

    This could be considered the 'main' function of the program. It is
    isolated here, though, to avoid having to deal with the command line
    or any other interface, so it can be imported and called easily from
    other programs.

    See the documentation of download_cms for a description of what
    output_dir, start_urls and remote_filters should be.

    language_codes should be a pair list in the format (code,extension),
    where code should be one ISO 639 language code, and extension can
    be any string, usually '.code', but you can use more exotic values.
    If you don't care about language codes, you should provide a list
    with a pair of empty strings.

    html_extension is the string that will be used to save HTML files,
    and it should include the dot, like `.html'.

    clean_html is a boolean value which tells the program to run
    mxTidy on the HTML input if it is malformed.
    """
    post_process_bundles = []
    html_parser.html_extension = html_extension
    # preprocess dictionary ugly trick
    for code, extension in language_codes:
        html_parser.extension_hook[code] = extension
    black_list_regex = create_mixed_expression(black_list)
    try:
        for lang_code, dummy in language_codes:
            post_process_data, log_params = download_cms(start_urls, output_dir,
                remote_filters, clean_html, lang_code, black_list_regex)
            post_process_bundles.append(post_process_data)
            generate_log(*log_params)
    except KeyboardInterrupt:
        print hr
        print " Aborted download. Files won't be processed correctly."
        print hr

    try:
        for html_files, visited_urls, clean_html in post_process_bundles:
            for local_file, url in html_files.iteritems():
                post_process_file(local_file, url, visited_urls, clean_html)
    except KeyboardInterrupt:
        print "Aborted post processing"

    # finally freeze all left temporary files
    try:
        for file, dummy in walk_local_files(".tmp", output_dir):
            new_name, old_ext = os.path.splitext(file)
            os.rename(file, new_name)
            print "Renamed %s to %s" % (file, new_name)
    except OSError, msg:
        print "Problems trying to freeze final files\n", msg


def usage_information(exit_code = SUCCESSFUL_EXIT, binary_name = ""):
    """Prints usage information and terminates execution."""
    if not binary_name:
        binary_name = sys.argv[0]
    print """Usage: %s [-hvc -d x -f x -l x -e x -b x] url_1 url_2 ... url_n

-h, --help
    Print this help screen.
-v, --version
    Print version number and exit.
-d xxx, --directory xxx
    Downloaded files will be put there. Defaults to `out'.
-f xxx, --filters xxx
    Whitespace separated list of URLs which will act as download
    filters. If you don't specify any, the filters will be set to
    the same values of the URLs you want to download, which is ok
    for basic recursive downloads.  Note that URLs you specify to
    download will be downloaded regardless of any filter.
-l xxx, --languages xxx
    A comma separated list of languages you want to download.
    Usually you will put here ISO 639 language codes. To each
    language you can add a colon and file intra extension which
    will be added to downloaded HTML files, otherwise the language
    code will be inserted as the intra extension preceded by a
    dot. Defaults to `en:.en,es:.es'.  You can also use an empty
    string, ignoring languages.
-e xxx, --html-extension xxx
    Set a specific filename extension for URLs which don't have
    one. Defaults to `.html'.
-c, --clean-html
    Use the mx.Tidy package to clean HTML if it's malformed.
    By default this option is disabled.
-b xxx, --black-list xxx
    Adds a regular expression to the blacklist of URLs which
    won't be downloaded. The regular expression uses Python's
    format and can match anywhere in the complete URL (see:
    http://python.org/doc/current/lib/re-syntax.html). Beware with
    metacharacters being escaped by the shell (like * or \\). You
    can repeat this parameter as much as you need.
    
Usage example:
 %s -d etxe -l "es:_c,eu:_e,en:_i" http://etxebide.efaber.net/
""" % (binary_name, binary_name)
    sys.exit(exit_code)


def process_command_line(argv = None):
    """Extracts from argv the options and returns them in a tuple.

    This function is a command line wrapper against main_process,
    it returns a tuple which you can `apply' calling main_process. If
    something in the command line is missing, the program will exit
    with a hopefully helpfull message.

    args should be a list with the full command line. If it is None
    or empty, the arguments will be extracted from sys.argv. The
    correct format of the accepted command line is documented by
    usage_information.
    """
    import getopt
    if not argv:
        argv = sys.argv

    short = "hvd:f:l:e:cb:"
    long = ["help", "version", "directory=", "filters=", "languages=",
        "html-extension=", "clean-html", "black-list="]

    try:
        opts, args = getopt.getopt(argv[1:], short, long)
    except getopt.GetoptError, msg:
        print "Error processing command line: %s\n" % msg
        usage_information(2)

    output_directory = "out/"
    remote_filters = []
    black_list = []
    language_codes = [("en", ".en"), ("es", ".es")]
    html_extension = ".html"
    clean_html = 0

    for option, value in opts:
        if option in ("-h", "--help"):
            usage_information()
        elif option in ("-v", "--version"):
            print HUMAN_VERSION
            sys.exit(SUCCESSFUL_EXIT)
        elif option in ("-d", "--directory"):
            output_directory = os.path.join(value, "")
        elif option in ("-f", "--filters"):
            remote_filters = value.split()
        elif option in ("-l", "--languages"):
            language_codes = []
            for language in value.split(","):
                split = language.split(":", 1)
                code = split[0]
                try:
                    extension = split[1]
                except IndexError:
                    extension = ".%s" % code
                language_codes.append((code,extension))
        elif option in ("-e", "--html-extension"):
            html_extension = value
        elif option in ("-c", "--clean-html"):
            clean_html = 1
        elif option in ("-b", "--black-list"):
            black_list.append(value)

    if not args:
        print "You have to provide at least one URL to download.\n"
        usage_information()

    if not remote_filters:
        remote_filters = args[:]

    # tell the user what options did (s)he choose
    print "Output directory will be `%s'" % output_directory
    print "Starting urls:",
    for f in args: print "`%s'" % f,
    print "\nRemote filters:",
    for f in remote_filters: print "`%s'" % f,
    print "\nLanguage codes (%d):" % len(language_codes),
    for f,g in language_codes: print "`%s:%s'" % (f,g),
    print "\nUsing extension: `%s'" % html_extension
    if not black_list: print "Not using blacklists"
    else:
        print "Expressions used to blacklist URLs:"
        for f in black_list: print "`%s'" % f,
    print
    return (output_directory, args, remote_filters, language_codes,
        html_extension, clean_html, black_list)


def create_mixed_expression(black_list):
    """Returns a single regex matching all those in black_list.

    Matching a certain pattern multiple times is useless. It's
    better to build a single regular expression which can match any
    possible pattern at the same time. Pass this function a list
    of regular expressions and a new regular expression will be
    returned which matches all of them at the same time (using OR).

    As a special case, if black_list is emtpy None is returned.

    This code was extracted from
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/81330.
    """
    if not black_list:
        return None

    return re.compile("(%s)" % "|".join(black_list))

if __name__ == "__main__":
    args = process_command_line()
    main_process(*args)
    print "Done"

